/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines involved with enforcing the rules of Mille Bournes.
 */

#include "mille.h"


static void calcStat(PLAYER *, STATUS);


/*
 * See if a player can play one of his cards.
 * If the card can be legally played on this turn, NULL is returned.
 * Otherwise, the address of an appropriate error message is returned.
 */
const char *
canPlayMsg(const PLAYER * pp, CARD card)
{
	int cardIndex;		/* index of card */
	static char buf[80];	/* message */

	const PLAYER * op = pp->mOpponent;

	/* make sure have a card */
	if (IsNoCard(card))
		return "No such card";

	cardIndex = -1;

	do
	{
		if (++cardIndex > HANDSIZE)
			return "Card not found";	/* card not in the hand */
	}
	while (pp->mCards[cardIndex] != card);

	if ((pp->mCardCount <= HANDSIZE) && (saveData.mCardsLeft > 0))
	{
		if (IsHazard(pp->mCoupCard) && (card == HazardToSafety(pp->mCoupCard)))
			return NULL;

		return "Must pick first";
	}

	/* can always play a safety card */
	if (IsSafety(card))
		return NULL;

	/* see if can use mileage card */
	if (IsMileage(card))
	{
		if (pp->mStopped)
		{
			if (IsRemedy(pp->mBattlePile) || IsNoCard(pp->mBattlePile) ||
				pp->mSafeCards[SafetyIndex(HazardToSafety(pp->mBattlePile))])
			{
					return "Need a Roll card";
			}

			sprintf(buf, "Need to remedy %s", cardNames[pp->mBattlePile]);

			return buf;
		}

		if (pp->mLimited && (card > TC_50MILE))
			return "Speed Limit is 50";

		if ((pp->mMileage + mileage[MileageIndex(card)]) > saveData.mMileageGoal)
		{
			sprintf(buf, "Exceeds goal of %d", saveData.mMileageGoal);

			return buf;
		}

		if ((card == TC_200MILE) &&
			(pp->mMileCards[MileageIndex(TC_200MILE)] >= 2))
		{
				return "Can only use two 200 cards";
		}

		return NULL;
	}

	/* see if we can use roll card */
	if (card == TC_ROLL)
	{
		if (!pp->mStopped)
			return "Not stopped";

		if (IsNoCard(pp->mBattlePile))
			return NULL;

		if ((pp->mBattlePile == TC_STOP) || IsRemedy(pp->mBattlePile))
			return NULL;

		if (pp->mSafeCards[SafetyIndex(HazardToSafety(pp->mBattlePile))])
			return NULL;

		sprintf(buf, "Need to remedy %s", cardNames[pp->mBattlePile]);

		return buf;
	}

	/* see if can end speed limit */
	if (card == TC_ENDOFLIMIT)
	{
		if (pp->mLimited)
			return NULL;

		return "No Speed Limit in effect";
	}

	/* see if can remedy current hazard */
	if (IsRemedy(card))
	{
		if (!pp->mStopped)
			return "Not stopped";

		if (!IsHazard(pp->mBattlePile))
			return "No hazard to remedy";

		if (pp->mBattlePile == RemedyToHazard(card))
			return NULL;

		sprintf(buf, "Wrong remedy for %s", cardNames[pp->mBattlePile]);

		return buf;
	}

	/* see if can play speed limit */
	if (card == TC_SPEEDLIMIT)
	{
		if (op->mSafeCards[SafetyIndex(TC_RIGHTOFWAY)])
			return "Protected by Right of Way";

		return NULL;
	}

	/* see if can play hazard */
	if (IsHazard(card))
	{
		if (op->mStopped && IsNoCard(op->mBattlePile))
			return "Opponent must play Roll";

		if (!op->mSafeCards[SafetyIndex(HazardToSafety(card))])
			return NULL;

		sprintf(buf, "Protected by %s", cardNames[HazardToSafety(card)]);

		return buf;
	}

	return "Unknown card";
}


/*
 * Play the specified card.
 * It is assumed that CanPlay was previously called to verify that the
 * use of the card is legal, and so legality checks are not made here.
 */
void
playCard(PLAYER * pp, CARD card)
{
	PLAYER * op = pp->mOpponent;
	CARD coupCard = pp->mCoupCard;

	discard(pp, card);

	/* Try to play mileage card */
	if (IsMileage(card))
	{
		pp->mMileage += mileage[MileageIndex(card)];
		pp->mMileCards[MileageIndex(card)]++;
		score(pp);

		return;
	}

	/* end of limit */
	if (card == TC_ENDOFLIMIT)
	{
		pp->mSpeedPile = card;
		pp->mLimited = FALSE;

		return;
	}

	/* remedy a hazard */
	if (IsRemedy(card))
	{
		pp->mBattlePile = card;
		checkStop(pp);

		return;
	}

	/* play a safety */
	if (IsSafety(card))
	{
		pp->mMoved = FALSE;	/* this is a free turn */
		pp->mSafeCards[SafetyIndex(card)] = TRUE;
		checkStop(pp);
		score(pp);

		/*
		 * Check for coup fourre
		 */
		if (IsNoCard(coupCard) || (card != HazardToSafety(coupCard)))
			return;

		pp->mCoupCards[SafetyIndex(card)] = TRUE;
		score(pp);

		if (coupCard == TC_SPEEDLIMIT)
		{
			pp->mSpeedPile = pp->mOldSpeed;

			return;
		}

		pp->mBattlePile = pp->mOldBattle;
		checkStop(pp);

		return;
	}

	/* play a speed limit on the opponent */
	if (card == TC_SPEEDLIMIT)
	{
		op->mOldSpeed = op->mSpeedPile;
		op->mSpeedPile = TC_SPEEDLIMIT;
		op->mCoupCard = TC_SPEEDLIMIT;
		op->mLimited = TRUE;

		return;
	}

	/* play a hazard on the opponent */
	if (IsHazard(card))
	{
		op->mOldBattle = op->mBattlePile;
		op->mBattlePile = card;
		op->mCoupCard = card;
		op->mStopped = TRUE;

		return;
	}
}


/*
 * Called when something happens which can affect start/stop of player.
 * Sets mStopped to indicate whether or not the player is stopped, and
 * mLimited to indicate whether or not speed limit is in effect.
 */
void
checkStop(PLAYER * pp)
{
	CARD card;
	int rightOfWay;

	rightOfWay = pp->mSafeCards[SafetyIndex(TC_RIGHTOFWAY)];

	pp->mLimited = ((pp->mSpeedPile == TC_SPEEDLIMIT) && !rightOfWay);
	card = pp->mBattlePile;

	if (card == TC_ROLL)
	{
		pp->mStopped = FALSE;

		return;
	}

	if (IsHazard(card) && !pp->mSafeCards[SafetyIndex(HazardToSafety(card))])
	{
		pp->mStopped = TRUE;

		return;
	}

	pp->mStopped = !rightOfWay;
}


/*
 * Determine the card needed to allow a player to play good mileage.
 * This is either the proper remedy card if the player is stopped,
 * or an end of limit card if a speed limit is in effect.
 */
CARD
fixCard(const PLAYER * pp)
{
	CARD card;		/* card to be fixed */

	if (!pp->mStopped)
	{
		/* need end of limit */
		if (pp->mLimited)
			return TC_ENDOFLIMIT;

		/* not stopped */
		return TC_NONE;
	}

	card = pp->mBattlePile;

	/* need to get a roll */
	if (!IsHazard(card) || (pp->mSafeCards[SafetyIndex(HazardToSafety(card))]))
		return TC_ROLL;

	/* need this remedy */
	return HazardToRemedy(card);
}


/*
 * Remove the specified card from the player's hand.
 * This simply turns it into a nonexistent card, and resorts the hand.
 */
void
discard(PLAYER * pp, CARD card)
{
	int cardIndex;	/* card to remove */

	pp->mMoved = TRUE;
	pp->mCoupCard = TC_NONE;
	pp->mPickIndex = -1;

	/* not a real card */
	if (IsNoCard(card))
		return;

	cardIndex = -1;

	do
	{
		/* card not in the hand */
		if (++cardIndex > HANDSIZE)
			return;
	}
	while (pp->mCards[cardIndex] != card);

	pp->mCards[cardIndex] = TC_NONE;
	pp->mCardCount--;

	saveData.mTypesLeft[card]--;

	sortHand(pp);
}


/*
 * Check to see if the hand is finished, and return status bits indicating
 * the result of the hand or game, from the point of view of the player.
 * Status is zero if the hand is not yet over.	The hand is over when
 * one player reaches the mileage goal and doesn't extend, or when we can
 * determine that no more points can be played for both players.
 */
STATUS
gameStatus(void)
{
	PLAYER * pp;		/* current player examined */
	STATUS	status;		/* status bits to return */
	STATUS	machStat;
	CARD	card;
	CARD	lowCard;
	BOOL	gotRoll;
	BOOL	gotEndLimit;
	BOOL	gotRemedy;
	int	score1;
	int	score2;
	int	i;

	status = 0;

	for (pp = saveData.mPlayers; pp < &saveData.mPlayers[2]; pp++)
	{
		if (pp->mMileage < saveData.mMileageGoal)
			continue;

		if ((saveData.mMileageGoal == S_SHORTGOAL) && pp->mExtendFunc(pp))
		{
			saveData.mMileageGoal = S_LONGGOAL;

			return 0;
		}

		pp->mTripDone = TRUE;
		score(pp);
		status |= ST_TRIPDONE;

		goto handover;
	}

	/* still have cards in deck */
	if (saveData.mCardsLeft > 0)
		return 0;

	for (pp = saveData.mPlayers; pp < &saveData.mPlayers[2]; pp++)
	{
		gotRoll = FALSE;
		gotEndLimit = FALSE;
		gotRemedy = FALSE;
		lowCard = 1000;

		for (i = 0; i < pp->mCardCount; i++)
		{
			card = pp->mCards[i];

			/* safeties can always be used */
			if (IsSafety(card))
				return 0;

			if (card == TC_ROLL)
				gotRoll = TRUE;

			if (card == TC_ENDOFLIMIT)
				gotEndLimit = TRUE;

			if (IsHazard(pp->mBattlePile) &&
				(card == HazardToRemedy(pp->mBattlePile)))
			{
				gotRemedy = TRUE;
			}

			if (IsMileage(card) && (card < lowCard))
				lowCard = card;
		}

		/* no mileage cards */
		if (!IsMileage(lowCard))
			continue;

		/* mileages too high for goal */
		if ((pp->mMileage + mileage[MileageIndex(lowCard)]) > saveData.mMileageGoal)
			continue;

		/* mileages too high for speed limit */
		if (pp->mLimited && !gotEndLimit && (lowCard > TC_50MILE))
			continue;

		/* can play a mileage card */
		if (!pp->mStopped)
			return 0;

		/* stopped and no roll or right of way */
		if (!gotRoll && !pp->mSafeCards[SafetyIndex(TC_RIGHTOFWAY)])
			continue;

		card = pp->mBattlePile;

		/* can remedy, roll, and do mileage */
		if (gotRemedy || IsNoCard(card) || IsRemedy(card))
			return 0;

		/* can use roll and do mileage */
		if (pp->mSafeCards[SafetyIndex(HazardToSafety(card))])
			return 0;
	}

	/*
	 * No progress can be made.  End the hand.
	 * Here when the hand is over.	See if the game is over too.
	 * Add in statistics for our convenience.
	 */
handover:
	status |= ST_HANDOVER;

	if (player->mScore > machine->mScore)
		status |= ST_HANDWON;

	if (player->mScore == machine->mScore)
		status |= ST_HANDTIED;

	saveData.mTotalHands++;
	score1 = player->mGameScore;
	score2 = machine->mGameScore;

	if (((score1 >= S_GAMEGOAL) || (score2 >= S_GAMEGOAL))
		&& (score1 != score2))
	{
		status |= ST_GAMEOVER;

		if (score1 > score2)
			status |= ST_GAMEWON;

		saveData.mTotalGames++;
	}

	calcStat(player, status);

	machStat = (status & ST_GAMEOVER);

	if ((status & ST_GAMEOVER) && ((status & ST_GAMEWON) == 0))
		machStat |= ST_GAMEWON;

	if ((status & (ST_HANDWON | ST_HANDTIED)) == 0)
		machStat |= ST_HANDWON;

	calcStat(machine, machStat);

	return status;
}


/*
 * Calculate statistics for a player.
 */
static void
calcStat(PLAYER * pp, STATUS status)
{
	int i;		/* index */
	int score;	/* our game score */

	if (pp->mBestHand < pp->mScore)
		pp->mBestHand = pp->mScore;

	if (status & ST_HANDWON)
	{
		pp->mHandsWon++;

		if ((status & ST_TRIPDONE) && (pp->mOpponent->mMileage == 0))
			pp->mShutouts++;
	}

	if (status & ST_GAMEOVER)
	{
		score = pp->mGameScore;

		for (i = 8; i >= 0; i--)
			pp->mTenGameScores[i + 1] = pp->mTenGameScores[i];

		pp->mTenGameScores[0] = score;
		pp->mAllGamesScore += score;

		if ((pp->mWorstGame > score) || (pp->mBestGame == 0))
			pp->mWorstGame = pp->mGameScore;

		if (pp->mBestGame < score)
			pp->mBestGame = score;

		if (status & ST_GAMEWON)
			pp->mGamesWon++;
	}
}


/*
 * Calculate the current score of a player.
 * The result is saved in the player structure.
 */
void
score(PLAYER * pp)
{
	int score = 0;
	int index;

	/* per safety */
	for (index = 0; index < SAFETYPES; index++)
	{
		if (pp->mSafeCards[index])
			score += S_SAFETY;
	}

	/* all four safeties */
	if (score == (SAFETYPES * S_SAFETY))
		score += S_ALLSAFETY;

	/* per coup fourre */
	for (index = 0; index < SAFETYPES; index++)
	{
		if (pp->mCoupCards[index])
			score += S_COUPFOURRE;
	}

	/* mileage played */
	score += pp->mMileage;

	/* done if not finished trip */
	if (!pp->mTripDone)
	{
		if (score > pp->mScore)
			pp->mGameScore += (score - pp->mScore);

		pp->mScore = score;

		return;
	}

	/* trip complete */
	score += S_TRIPDONE;

	/* delayed action */
	if (saveData.mCardsLeft == 0)
		score += S_DELAYED;

	/* safe trip */
	if (pp->mMileCards[MileageIndex(TC_200MILE)] == 0)
		score += S_SAFETRIP;

	/* extension */
	if (saveData.mMileageGoal == S_LONGGOAL)
		score += S_EXTENSION;

	/* shut out */
	if (pp->mOpponent->mMileage == 0)
		score += S_SHUTOUT;

	if (score > pp->mScore)
		pp->mGameScore += (score - pp->mScore);

	pp->mScore = score;
}


/*
 * Try to add the specified number of cards to the player's hand.
 * The cards are added to the end of the hand, and then resorted.
 * Returns error string if no picking is impossible, or NULL if at
 * least one card can be drawn.
 */
const char *
pickCards(PLAYER * pp, int count)
{
	int num;		/* number of cards actually got */

	pp->mCoupCard = TC_NONE;

	for (num = 0; num < count; num++)
	{
		if (saveData.mCardsLeft <= 0)
		{
			if (num)
				break;

			return "No cards left in deck";
		}

		if ((pp->mCardCount > HANDSIZE)
			|| (IsCard(pp->mCards[HANDSIZE])))
		{
			if (num)
				break;

			return "Your hand is full";
		}

		pp->mCards[HANDSIZE] = saveData.mCardDeck[--saveData.mCardsLeft];
		pp->mCardCount++;
		pp->mPickIndex = HANDSIZE;

		sortHand(pp);
	}

	return NULL;
}


/*
 * Sort a player's hand so that cards of the same type appear together.
 * The card just picked is also sorted.	The rankings are such that
 * blank cards are placed at the end.
 */
void
sortHand(PLAYER * pp)
{
	int	i;
	int	j;
	CARD *	t1;
	CARD *	t2;
	CARD	tmp;

	for (i = HANDSIZE - 1; i >= 0; i--)
	{
		for (j = 0; j <= i; j++)
		{
			t1 = &pp->mCards[j];
			t2 = &pp->mCards[j+1];

			if (cardRanks[*t1] > cardRanks[*t2])
			{
				tmp = *t1;
				*t1 = *t2;
				*t2 = tmp;

				if ((pp->mPickIndex == j) || (pp->mPickIndex == (j+1)))
					pp->mPickIndex = (2 * j + 1) - pp->mPickIndex;
			}
		}
	}
}

/* END CODE */
